import os
from collections import ChainMap
from pathlib import Path
from typing import Any, Dict

from ramalama.common import container_manager, default_image
from ramalama.toml_parser import TOMLParser

DEFAULT_PORT_RANGE = (8080, 8090)


def get_store():
    if os.geteuid() == 0:
        return "/var/lib/ramalama"

    return os.path.expanduser("~/.local/share/ramalama")


def use_container():
    use_container = os.getenv("RAMALAMA_IN_CONTAINER")
    if use_container:
        return use_container.lower() == "true"

    conman = container_manager()
    return conman is not None


def load_config() -> Dict[str, Any]:
    """Load configuration from a list of paths, in priority order."""
    parser = TOMLParser()
    config_path = os.getenv("RAMALAMA_CONFIG")
    if config_path:
        return parser.parse_file(config_path)

    config = {}
    config_paths = [
        "/usr/share/ramalama/ramalama.conf",
        "/usr/local/share/ramalama/ramalama.conf",
        "/etc/ramalama/ramalama.conf",
    ]
    config_home = os.getenv("XDG_CONFIG_HOME", os.path.join("~", ".config"))
    config_paths.extend([os.path.expanduser(os.path.join(config_home, "ramalama", "ramalama.conf"))])

    # Load configuration from each path
    for path in config_paths:
        if os.path.exists(path):
            # Load the main config file
            config = parser.parse_file(path)
        if os.path.isdir(path + ".d"):
            # Load all .conf files in ramalama.conf.d directory
            for conf_file in sorted(Path(path + ".d").glob("*.conf")):
                config = parser.parse_file(conf_file)

    return config


def load_config_from_env(config: Dict[str, Any], env: Dict):
    """Load configuration from environment variables."""
    envvars = {
        'container': 'RAMALAMA_IN_CONTAINER',
        'engine': 'RAMALAMA_CONTAINER_ENGINE',
        'image': 'RAMALAMA_IMAGE',
        'store': 'RAMALAMA_STORE',
        'transport': 'RAMALAMA_TRANSPORT',
    }
    for k, v in envvars.items():
        if value := env.get(v):
            config[k] = value


def int_tuple_as_str(input: tuple) -> str:
    return '-'.join(map(str, input))


def load_config_defaults(config: Dict[str, Any]):
    """Set configuration defaults if these are not yet set."""
    config.setdefault('carimage', "registry.access.redhat.com/ubi9-micro:latest")
    config.setdefault('container', use_container())
    config.setdefault('ctx_size', 2048)
    config.setdefault('engine', container_manager())
    config.setdefault('env', [])
    config.setdefault('host', "0.0.0.0")
    config.setdefault('image', default_image())
    config.setdefault(
        'images',
        {
            "ASAHI_VISIBLE_DEVICES": "quay.io/ramalama/asahi",
            "ASCEND_VISIBLE_DEVICES": "quay.io/ramalama/cann",
            "CUDA_VISIBLE_DEVICES": "quay.io/ramalama/cuda",
            "HIP_VISIBLE_DEVICES": "quay.io/ramalama/rocm",
            "INTEL_VISIBLE_DEVICES": "quay.io/ramalama/intel-gpu",
        },
    )
    config.setdefault('keep_groups', False)
    config.setdefault('ngl', -1)
    config.setdefault('threads', -1)
    config.setdefault('nocontainer', False)
    config.setdefault('port', int_tuple_as_str(DEFAULT_PORT_RANGE))
    config.setdefault('pull', "newer")
    config.setdefault('runtime', 'llama.cpp')
    config.setdefault('store', get_store())
    config.setdefault('temp', "0.8")
    config.setdefault('transport', "ollama")
    config.setdefault('use_model_store', True)


class Config(ChainMap):
    def __init__(self, from_env, from_file, default):
        super().__init__(from_env, from_file, default)

    @property
    def from_env(self):
        return self.maps[0]

    @property
    def from_file(self):
        return self.maps[1]

    @property
    def default(self):
        return self.maps[2]

    def is_set(self, key):
        if key in self.from_env:
            return True
        if key in self.from_file:
            return True
        return False


def load_and_merge_config() -> Dict[str, Any]:
    """Load configuration from files, merge with environment variables and set defaults for options not set."""
    config = load_config()

    ramalama_config = config.setdefault('ramalama', {})

    env_config = {}
    load_config_from_env(env_config, os.environ)

    default_config = {}
    load_config_defaults(default_config)

    return Config(env_config, ramalama_config, default_config)


CONFIG = load_and_merge_config()



"""Version of RamaLamaPy."""


def version():
    version = "0.8.0"
    return version


def print_version(args):
    if args.quiet:
        print(version())
    else:
        print("ramalama version %s" % version())



import glob
import json
import os
import subprocess
import sys

from ramalama.common import check_nvidia, exec_cmd, get_accel_env_vars, perror, podman_machine_accel, run_cmd
from ramalama.console import EMOJI


class Engine:

    def __init__(self, args):
        self.exec_args = [
            args.engine,
            "run",
            "--rm",
        ]
        self.use_docker = os.path.basename(args.engine) == "docker"
        self.use_podman = os.path.basename(args.engine) == "podman"
        self.args = args
        self.add_container_labels()
        self.add_device_options()
        self.add_env_option()
        self.add_network()
        self.add_oci_runtime()
        self.add_port_option()
        self.add_privileged_options()
        self.add_pull_newer()
        self.add_rag()
        self.add_subcommand_env()
        self.add_tty_option()
        self.handle_podman_specifics()
        self.add_detach_option()
        self.debug = args.debug

    def add_label(self, label):
        self.add(["--label", label])

    def add_container_labels(self):
        label_map = {
            "MODEL": "ai.ramalama.model",
            "engine": "ai.ramalama.engine",
            "runtime": "ai.ramalama.runtime",
            "port": "ai.ramalama.port",
            "subcommand": "ai.ramalama.command",
        }
        for arg, label_prefix in label_map.items():
            if hasattr(self.args, arg):
                value = getattr(self.args, arg)
                if value:
                    self.add_label(f"{label_prefix}={value}")

    def add_pull_newer(self):
        if not self.args.dryrun and self.use_docker and self.args.pull == "newer":
            try:
                if not self.args.quiet:
                    print(f"Checking for newer image {self.args.image}")
                run_cmd([self.args.engine, "pull", "-q", self.args.image], ignore_all=True)
            except Exception:  # Ignore errors, the run command will handle it.
                pass
        else:
            self.exec_args += ["--pull", self.args.pull]

    def add_network(self):
        if hasattr(self.args, "network") and self.args.network:
            self.exec_args += ["--network", self.args.network]

    def add_oci_runtime(self):
        if hasattr(self.args, "oci_runtime") and self.args.oci_runtime:
            self.exec_args += ["--runtime", self.args.oci_runtime]
            return
        if check_nvidia() == "cuda":
            if self.use_docker:
                self.exec_args += ["--runtime", "nvidia"]
            else:
                self.exec_args += ["--runtime", "/usr/bin/nvidia-container-runtime"]

    def add_privileged_options(self):
        if hasattr(self.args, "privileged") and self.args.privileged:
            self.exec_args += ["--privileged"]
        else:
            self.exec_args += [
                "--security-opt=label=disable",
            ]
            if not hasattr(self.args, "nocapdrop"):
                self.exec_args += [
                    "--cap-drop=all",
                    "--security-opt=no-new-privileges",
                ]

    def add_subcommand_env(self):
        if EMOJI and hasattr(self.args, "subcommand") and self.args.subcommand == "run":
            if os.path.basename(self.args.engine) == "podman":
                self.exec_args += ["--env", "LLAMA_PROMPT_PREFIX=🦭 > "]
            if self.use_docker:
                self.exec_args += ["--env", "LLAMA_PROMPT_PREFIX=🐋 > "]

    def add_env_option(self):
        if hasattr(self.args, "env"):
            for env in self.args.env:
                self.exec_args += ["--env", env]

    def add_tty_option(self):
        if sys.stdout.isatty() or sys.stdin.isatty():
            self.exec_args += ["-t"]

    def add_detach_option(self):
        if hasattr(self.args, "detach") and self.args.detach is True:
            self.exec_args += ["-d"]

    def add_port_option(self):
        if hasattr(self.args, "port"):
            self.exec_args += ["-p", f"{self.args.port}:{self.args.port}"]

    def add_device_options(self):
        if hasattr(self.args, "device") and self.args.device:
            for device_arg in self.args.device:
                self.exec_args += ["--device", device_arg]

        if podman_machine_accel:
            self.exec_args += ["--device", "/dev/dri"]

        for path in ["/dev/dri", "/dev/kfd", "/dev/accel", "/dev/davinci*", "/dev/devmm_svm", "/dev/hisi_hdc"]:
            for dev in glob.glob(path):
                self.exec_args += ["--device", dev]

        for k, v in get_accel_env_vars().items():
            # Special case for Cuda
            if k == "CUDA_VISIBLE_DEVICES":
                if self.use_docker:
                    self.exec_args += ["--gpus", "all"]
                else:
                    # newer Podman versions support --gpus=all, but < 5.0 do not
                    self.exec_args += ["--device", "nvidia.com/gpu=all"]

            self.exec_args += ["-e", f"{k}={v}"]

    def add_rag(self):
        if not hasattr(self.args, "rag") or not self.args.rag:
            return

        if os.path.exists(self.args.rag):
            rag = os.path.realpath(self.args.rag)
            # Added temp read write because vector database requires write access even if nothing is written
            self.exec_args.append(f"--mount=type=bind,source={rag},destination=/rag/vector.db,rw=true")
        else:
            self.exec_args.append(f"--mount=type=image,source={self.args.rag},destination=/rag,rw=true")

    def handle_podman_specifics(self):
        if hasattr(self.args, "podman_keep_groups") and self.args.podman_keep_groups:
            self.exec_args += ["--group-add", "keep-groups"]

    def add(self, newargs):
        self.exec_args += newargs

    def dryrun(self):
        dry_run(self.exec_args)

    def run(self):
        run_cmd(self.exec_args, debug=self.debug)

    def exec(self):
        exec_cmd(self.exec_args, debug=self.debug)


def dry_run(args):
    for arg in args:
        if not arg:
            continue
        if " " in arg:
            print('"%s"' % arg, end=" ")
        else:
            print("%s" % arg, end=" ")
    print()


def images(args):
    conman = args.engine
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "images"]
    if hasattr(args, "noheading") and args.noheading:
        conman_args += ["--noheading"]

    if hasattr(args, "notrunc") and args.notrunc:
        conman_args += ["--no-trunc"]

    if args.format:
        conman_args += [f"--format={args.format}"]

    try:
        output = run_cmd(conman_args, debug=args.debug).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return output.split("\n")
    except subprocess.CalledProcessError as e:
        perror("ramalama list command requires a running container engine")
        raise (e)


def containers(args):
    conman = args.engine
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "ps", "-a", "--filter", "label=ai.ramalama"]
    if hasattr(args, "noheading") and args.noheading:
        conman_args += ["--noheading"]

    if hasattr(args, "notrunc") and args.notrunc:
        conman_args += ["--no-trunc"]

    if args.format:
        conman_args += [f"--format={args.format}"]

    try:
        output = run_cmd(conman_args, debug=args.debug).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return output.split("\n")
    except subprocess.CalledProcessError as e:
        perror("ramalama list command requires a running container engine")
        raise (e)


def info(args):
    conman = args.engine
    if conman == "":
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "info", "--format", "json"]
    try:
        output = run_cmd(conman_args, debug=args.debug).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return json.loads(output)
    except FileNotFoundError as e:
        return str(e)


def stop_container(args, name):
    if not name:
        raise ValueError("must specify a container name")
    conman = args.engine
    if conman == "":
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "stop", "-t=0"]
    ignore_stderr = False
    if args.ignore:
        if conman == "podman":
            conman_args += ["--ignore", str(args.ignore)]
        else:
            ignore_stderr = True

    conman_args += [name]
    try:
        run_cmd(conman_args, ignore_stderr=ignore_stderr, debug=args.debug)
    except subprocess.CalledProcessError:
        if args.ignore and conman == "docker":
            return
        else:
            raise



import os
import shutil
import subprocess
import tempfile
from urllib.parse import urlparse

from ramalama.common import accel_image, download_file, get_accel_env_vars, run_cmd, set_accel_env_vars
from ramalama.config import CONFIG
from ramalama.engine import Engine


class Rag:
    model = ""
    target = ""

    def __init__(self, target):
        if not target.islower():
            raise ValueError(f"invalid reference format: repository name '{target}' must be lowercase")
        self.target = target
        set_accel_env_vars()

    def build(self, source, target, args):
        print(f"\nBuilding {target}...")
        contextdir = os.path.dirname(source)
        src = os.path.basename(source)
        print(f"adding {src}...")
        cfile = f"""\
FROM scratch
COPY {src} /vector.db
"""
        containerfile = tempfile.NamedTemporaryFile(dir=source)
        # Open the file for writing.
        with open(containerfile.name, 'w') as c:
            c.write(cfile)
        if args.debug:
            print(f"\nContainerfile: {containerfile.name}\n{cfile}")
        imageid = (
            run_cmd(
                [
                    args.engine,
                    "build",
                    "--no-cache",
                    "--network=none",
                    "-q",
                    "-t",
                    target,
                    "-f",
                    containerfile.name,
                    contextdir,
                ],
                debug=args.debug,
            )
            .stdout.decode("utf-8")
            .strip()
        )
        return imageid

    def _handle_docs_path(self, path, docs_path, exec_args):
        """Adds a volume mount if path exists, otherwise downloads from URL."""
        if os.path.exists(path):
            fpath = os.path.realpath(path)
            self.engine.add(["-v", f"{fpath}:/docs/{fpath}:ro,z"])
            return False
        try:
            parsed = urlparse(path)
            if parsed.scheme == "" or parsed.path == "":
                raise ValueError(f"{path} does not exist")
            dpath = docs_path + parsed.path
            os.makedirs(os.path.dirname(dpath), exist_ok=True)
            download_file(path, dpath)
            return True  # docsdb was used
        except RuntimeError as e:
            shutil.rmtree(docs_path, ignore_errors=True)
            raise e

    def generate(self, args):
        args.nocapdrop = True
        self.engine = Engine(args)
        # force accel_image to use -rag version. Drop TAG if it exists
        # so that accel_image will add -rag to the image specification.
        args.rag = "rag"
        args.image = args.image.split(":")[0]
        args.image = accel_image(CONFIG, args)

        if not args.container:
            raise KeyError("rag command requires a container. Can not be run with --nocontainer option.")
        if not args.engine or args.engine == "":
            raise KeyError("rag command requires a container. Can not be run without a container engine.")

        tmpdir = "."
        if not os.access(tmpdir, os.W_OK):
            tmpdir = "/tmp"

        docsdb = tempfile.TemporaryDirectory(dir=tmpdir, prefix='RamaLama_docs_')
        docsdb_used = False

        for path in args.PATH:
            if self._handle_docs_path(path, docsdb.name, self.engine.exec_args):
                docsdb_used = True

        if docsdb_used:
            self.engine.add(["-v", f"{docsdb.name}:/docs/{docsdb.name}:ro,Z"])

        ragdb = tempfile.TemporaryDirectory(dir=tmpdir, prefix='RamaLama_rag_')
        dbdir = os.path.join(ragdb.name, "vectordb")
        os.mkdir(dbdir)
        self.engine.add(["-v", f"{dbdir}:/output:z"])
        for k, v in get_accel_env_vars().items():
            # Special case for Cuda
            if k == "CUDA_VISIBLE_DEVICES":
                if os.path.basename(args.engine) == "docker":
                    self.engine.add(["--gpus", "all"])
                else:
                    # newer Podman versions support --gpus=all, but < 5.0 do not
                    self.engine.add(["--device", "nvidia.com/gpu=all"])

            self.engine.add(["-e", f"{k}={v}"])

        self.engine.add([args.image])
        self.engine.add(["doc2rag", "/output", "/docs/"])
        try:
            self.engine.run()
            print(self.build(dbdir, self.target, args))
        except subprocess.CalledProcessError as e:
            raise e
        finally:
            shutil.rmtree(ragdb.name, ignore_errors=True)
            shutil.rmtree(docsdb.name, ignore_errors=True)
